# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import uuid

from oslo_log import log as logging

from shaker.engine.aggregators import base


LOG = logging.getLogger(__name__)


def mean(array):
    if not array:
        return 0
    array = [x for x in array if x]
    return sum(array) / len(array)


def safe_max(array):
    return max(x for x in array if x)


def safe_min(array):
    return min(x for x in array if x)


class TrafficAggregator(base.BaseAggregator):
    def __init__(self, test_definition):
        super(TrafficAggregator, self).__init__(test_definition)

    def test_summary(self, test_data):
        chart = []
        xs = []
        mean_v = collections.defaultdict(list)

        for iteration in test_data['results_per_iteration']:
            xs.append(len(iteration['results_per_agent']))
            for k, v in iteration['stats'].items():
                mean_v[k].append(v['mean'])

        for k in mean_v.keys():
            chart.append(['Mean %s' % k] + mean_v[k])

        chart.append(['x'] + xs)
        test_data.update({
            'chart': chart,
        })

    def iteration_summary(self, iteration_data):
        max_v = collections.defaultdict(list)
        min_v = collections.defaultdict(list)
        mean_v = collections.defaultdict(list)
        unit_v = dict()
        chart = []

        nodes = []
        for one in iteration_data['results_per_agent']:
            nodes.append(one['agent']['node'])
            chart += one['chart']

            for k, v in one['stats'].items():
                max_v[k].append(v['max'])
                min_v[k].append(v['min'])
                mean_v[k].append(v['mean'])
                unit_v[k] = v['unit']

        stats = {}
        node_chart = [['x'] + nodes]

        for k in max_v.keys():
            stats[k] = dict(max=max(max_v[k]),
                            min=min(min_v[k]),
                            mean=mean(mean_v[k]),
                            unit=unit_v[k])
            node_chart.append(['Mean %s' % k] + mean_v[k])
            node_chart.append(['Max %s' % k] + max_v[k])
            node_chart.append(['Min %s' % k] + min_v[k])

        iteration_data.update({
            'uuid': uuid.uuid4(),
            'stats': stats,
            'x-chart': chart,
            'node_chart': node_chart,
        })

    def agent_summary(self, agent_data):
        # convert bps to Mbps
        for idx, item_meta in enumerate(agent_data['meta']):
            if item_meta[1] == 'bps':
                for row in agent_data['samples']:
                    if row[idx]:
                        row[idx] = float(row[idx]) / 1024 / 1024
                item_meta[1] = 'Mbps'

        # calculate stats
        agent_data['stats'] = dict()
        agent_data['chart'] = []

        for idx, item_meta in enumerate(agent_data['meta']):
            column = [row[idx] for row in agent_data['samples']]

            item_title = item_meta[0]
            if item_title != 'time':
                agent_data['stats'][item_title] = {
                    'max': safe_max(column),
                    'min': safe_min(column),
                    'mean': mean(column),
                    'unit': item_meta[1],
                }
            agent_data['chart'].append([item_title] + column)

        # drop stdout
        del agent_data['stdout']



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import os
import time
import uuid

from oslo_config import cfg
from oslo_log import log as logging
import yaml

from shaker.engine import config
from shaker.engine import deploy
from shaker.engine import executors as executors_classes
from shaker.engine import messaging
from shaker.engine import report
from shaker.engine import utils


LOG = logging.getLogger(__name__)


class Quorum(object):
    def __init__(self, message_queue, polling_interval):
        self.message_queue = message_queue
        self.polling_interval = polling_interval

    def wait_join(self, agent_ids):
        agent_ids = set(agent_ids)
        LOG.debug('Waiting for quorum of agents: %s', agent_ids)
        alive_agents = set()
        for message, reply_handler in self.message_queue:
            msg_agent_id = message.get('agent_id')

            if msg_agent_id not in agent_ids:
                reply_handler(dict(operation='none'))
                continue

            alive_agents.add(msg_agent_id)

            reply_handler(dict(operation='configure',
                               polling_interval=self.polling_interval))

            LOG.debug('Alive agents: %s', alive_agents)

            if alive_agents == agent_ids:
                LOG.info('All expected agents are alive')
                break

    def run_test_case(self, test_case):
        working_agents = set()
        replied_agents = set()
        result = {}

        start_at = int(time.time()) + self.polling_interval * 2

        for message, reply_handler in self.message_queue:
            agent_id = message.get('agent_id')
            operation = message.get('operation')

            reply = {'operation': 'none'}

            if agent_id not in test_case:
                reply_handler(reply)
                continue

            # message from a known agent
            test = test_case[agent_id]

            if operation == 'poll':
                reply = {
                    'operation': 'execute',
                    'start_at': start_at,
                    'command': test.get_command(),
                }
                working_agents.add(agent_id)
            elif operation == 'reply':
                replied_agents.add(agent_id)
                result[agent_id] = test.process_reply(message)

            reply_handler(reply)

            LOG.debug('Working agents: %s', working_agents)
            LOG.debug('Replied agents: %s', replied_agents)

            if replied_agents >= set(test_case.keys()):
                LOG.info('Received all replies for test case: %s', test_case)
                break

        return result


def read_scenario():
    scenario_raw = utils.read_file(cfg.CONF.scenario)
    scenario = yaml.safe_load(scenario_raw)
    scenario['file_name'] = cfg.CONF.scenario
    LOG.debug('Scenario: %s', scenario)
    return scenario


def _extend_agents(agents):
    for agent in agents.values():
        if agent.get('slave_id'):
            agent['slave'] = copy.deepcopy(agents[agent['slave_id']])
        if agent.get('master_id'):
            agent['master'] = copy.deepcopy(agents[agent['master_id']])


def _pick_agents(agents, size):
    # slave agents do not execute any tests
    agents = [a for a in agents.values() if a.get('mode') != 'slave']

    if not size or size == 'full':
        yield agents
    elif size == 'linear_progression':
        for i in range(len(agents)):
            yield agents[:i + 1]
    elif size == 'quadratic_progression':
        n = len(agents)
        seq = [n]
        while n > 1:
            n //= 2
            seq.append(n)
        seq.reverse()
        for i in seq:
            yield agents[:i]


def execute(quorum, execution, agents):
    _extend_agents(agents)

    result = []

    for test in execution['tests']:
        LOG.debug('Running test %s on all agents', test)

        results_per_iteration = []
        for selected_agents in _pick_agents(agents, execution.get('size')):
            executors = dict((a['id'], executors_classes.get_executor(test, a))
                             for a in selected_agents)

            test_case_result = quorum.run_test_case(executors)
            values = test_case_result.values()
            for v in values:
                v['uuid'] = str(uuid.uuid4())
            results_per_iteration.append({
                'agents': selected_agents,
                'results_per_agent': values,
            })

        test['uuid'] = str(uuid.uuid4())
        result.append({
            'results_per_iteration': results_per_iteration,
            'definition': test,
        })

    LOG.info('Execution is done')
    return result


def main():
    utils.init_config_and_logging(
        config.COMMON_OPTS + config.OPENSTACK_OPTS + config.SERVER_OPTS +
        config.REPORT_OPTS
    )

    scenario = read_scenario()

    deployment = None
    agents = {}
    result = []

    try:
        deployment = deploy.Deployment(cfg.CONF.server_endpoint)

        if (cfg.CONF.os_username and cfg.CONF.os_password and
                cfg.CONF.os_tenant_name and cfg.CONF.os_auth_url):
            deployment.connect_to_openstack(
                cfg.CONF.os_username, cfg.CONF.os_password,
                cfg.CONF.os_tenant_name, cfg.CONF.os_auth_url,
                cfg.CONF.os_region_name, cfg.CONF.external_net,
                cfg.CONF.flavor_name, cfg.CONF.image_name)

        agents = deployment.deploy(scenario['deployment'],
                                   base_dir=os.path.dirname(cfg.CONF.scenario))
        LOG.debug('Deployed agents: %s', agents)

        if not agents:
            LOG.warning('No agents deployed.')
        else:
            message_queue = messaging.MessageQueue(cfg.CONF.server_endpoint)

            quorum = Quorum(message_queue, cfg.CONF.polling_interval)
            quorum.wait_join(set(agents.keys()))

            result = execute(quorum, scenario['execution'], agents)
            LOG.debug('Result: %s', result)
    except Exception as e:
        LOG.error('Error while executing scenario: %s', e)
    finally:
        if deployment:
            deployment.cleanup()

    report_data = dict(scenario=scenario,
                       agents=agents.values(),
                       result=result)
    if cfg.CONF.output:
        utils.write_file(json.dumps(report_data), cfg.CONF.output)

    report.generate_report(cfg.CONF.report_template, cfg.CONF.report,
                           report_data)


if __name__ == "__main__":
    main()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import csv

from shaker.openstack.common import log as logging


LOG = logging.getLogger(__name__)


class CommandLine(object):
    def __init__(self, command):
        self.commands = [command]

    def add(self, param_name, param_value=None):
        self.commands.append('%s' % param_name)
        if param_value:
            self.commands.append(str(param_value))

    def make(self):
        return ' '.join(self.commands)


class BaseExecutor(object):
    def __init__(self, test_definition, agent):
        super(BaseExecutor, self).__init__()
        self.test_definition = test_definition
        self.agent = agent

    def get_command(self):
        return None

    def process_reply(self, message):
        LOG.debug('Test %s on agent %s finished with %s',
                  self.test_definition, self.agent, message)
        return dict(stdout=message.get('stdout'),
                    stderr=message.get('stderr'),
                    command=self.get_command(),
                    agent=self.agent)


class ShellExecutor(BaseExecutor):
    def get_command(self):
        return self.test_definition['method']


class NetperfExecutor(BaseExecutor):
    def get_command(self):
        cmd = CommandLine('netperf')
        cmd.add('-H', self.agent['slave']['ip'])
        cmd.add('-l', self.test_definition.get('time') or 60)
        cmd.add('-t', self.test_definition.get('method') or 'TCP_STREAM')
        return cmd.make()


class NetperfWrapperExecutor(BaseExecutor):
    def get_command(self):
        target_ip = self.agent['slave']['ip']
        return ('netperf-wrapper -H %(ip)s -f stats %(method)s' %
                dict(ip=target_ip,
                     method=self.test_definition['method']))


class IperfExecutor(BaseExecutor):
    def get_command(self):
        cmd = CommandLine('sudo nice -n -20 iperf')
        cmd.add('--client', self.agent['slave']['ip'])
        cmd.add('--format', 'm')
        cmd.add('--nodelay')
        if self.test_definition.get('mss'):
            cmd.add('--mss', self.test_definition.get('mss'))
        cmd.add('--len', self.test_definition.get('buffer_size') or '8k')
        if self.test_definition.get('udp'):
            cmd.add('--udp')
        cmd.add('--time', self.test_definition.get('time') or 60)
        cmd.add('--parallel', self.test_definition.get('threads') or 1)
        if self.test_definition.get('csv'):
            cmd.add('--reportstyle', 'C')
        if self.test_definition.get('interval'):
            cmd.add('--interval', self.test_definition.get('interval'))
        return cmd.make()


def _calc_stats(array):
    return dict(max=max(array), min=min(array), avg=sum(array) / len(array))


class IperfGraphExecutor(IperfExecutor):
    def get_command(self):
        self.test_definition['csv'] = True
        self.test_definition['interval'] = '1'
        return super(IperfGraphExecutor, self).get_command()

    def process_reply(self, message):
        result = super(IperfGraphExecutor, self).process_reply(message)

        samples = collections.defaultdict(list)
        streams = {}
        stream_count = 0

        for row in csv.reader(result['stdout'].split('\n')):
            if row:
                thread = row[5]
                if thread not in streams:
                    streams[thread] = stream_count
                    stream_count += 1

                samples['time'].append(float(row[6].split('-')[1]))
                samples['bandwidth_%s' % streams[thread]].append(
                    float(row[8]) / 1024 / 1024)

        # the last line is summary, remove its items
        for arr in samples.values():
            arr.pop()

        result['samples'] = samples

        # todo calculate stats correctly for multiple threads
        for stream in streams.values():
            result['stats'] = _calc_stats(
                samples['bandwidth_%s' % stream])

        return result


EXECUTORS = {
    'shell': ShellExecutor,
    'netperf': NetperfExecutor,
    'iperf': IperfExecutor,
    'iperf_graph': IperfGraphExecutor,
    'netperf_wrapper': NetperfWrapperExecutor,
    '_default': ShellExecutor,
}


def get_executor(test_definition, agent):
    # returns executor of the specified test on the specified agent
    executor_class = test_definition['class']
    klazz = EXECUTORS.get(executor_class, EXECUTORS['_default'])
    return klazz(test_definition, agent)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from novaclient import client as nova_client_pkg
from oslo_log import log as logging


LOG = logging.getLogger(__name__)

NOVA_VERSION = '2'


def create_client(keystone_session, os_region_name):
    return nova_client_pkg.Client(NOVA_VERSION, session=keystone_session,
                                  region_name=os_region_name)


def get_available_compute_nodes(nova_client):
        return [svc.host
                for svc in nova_client.services.list(binary='nova-compute')
                if svc.state == 'up']


def is_flavor_exists(nova_client, flavor_name):
    for flavor in nova_client.flavors.list():
        if flavor.to_dict()['name'] == flavor_name:
            return True
    return False


def _poll_for_status(poll_fn, obj_id, final_ok_states, poll_period=20,
                     status_field="status"):
    LOG.debug('Poll server %(id)s, waiting for any of statuses %(statuses)s',
              dict(id=obj_id, statuses=final_ok_states))
    while True:
        obj = poll_fn(obj_id)
        status = getattr(obj, status_field)
        if status:
            status = status.lower()

        LOG.debug('Server %(id)s has status %(status)s',
                  dict(id=obj_id, status=status))

        if status in final_ok_states:
            break
        elif status == "error":
            raise Exception(obj.fault['message'])

        time.sleep(poll_period)


def wait_server_shutdown(nova_client, server_id):
    _poll_for_status(nova_client.servers.get, server_id, ['shutoff'])


def wait_server_snapshot(nova_client, server_id):
    task_state_field = "OS-EXT-STS:task_state"
    server = nova_client.servers.get(server_id)
    if hasattr(server, task_state_field):
        _poll_for_status(nova_client.servers.get, server.id, [None, '-', ''],
                         status_field=task_state_field)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv

from shaker.engine.executors import base


class IperfExecutor(base.BaseExecutor):
    def get_command(self):
        cmd = base.CommandLine('sudo nice -n -20 iperf')
        cmd.add('--client', self.agent['slave']['ip'])
        cmd.add('--format', 'm')
        cmd.add('--nodelay')
        if self.test_definition.get('mss'):
            cmd.add('--mss', self.test_definition.get('mss'))
        cmd.add('--len', self.test_definition.get('buffer_size') or '8k')
        if self.test_definition.get('udp'):
            cmd.add('--udp')
            if self.test_definition.get('bandwidth'):
                cmd.add('--bandwidth', self.test_definition.get('bandwidth'))
        cmd.add('--time', self.test_definition.get('time') or 60)
        cmd.add('--parallel', self.test_definition.get('threads') or 1)
        if self.test_definition.get('csv'):
            cmd.add('--reportstyle', 'C')
        if self.test_definition.get('interval'):
            cmd.add('--interval', self.test_definition.get('interval'))
        return cmd.make()


class IperfGraphExecutor(IperfExecutor):
    def get_command(self):
        self.test_definition['csv'] = True
        self.test_definition['interval'] = 1
        return super(IperfGraphExecutor, self).get_command()

    def process_reply(self, message):
        result = super(IperfGraphExecutor, self).process_reply(message)

        samples = []
        threads_count = self.test_definition.get('threads') or 1

        for row in csv.reader(result['stdout'].split('\n')):
            if row and len(row) > 8:
                thread = row[5]
                if threads_count > 1 and thread != -1:
                    # ignore individual per-thread data
                    continue

                start, end = row[6].split('-')
                samples.append([float(end), int(row[8])])

        samples.pop()  # the last line is summary, remove it

        result['samples'] = samples
        result['meta'] = [['time', 's'], ['bandwidth', 'bps']]
        return result



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from oslo_config import cfg
from shaker.engine import utils


COMMON_OPTS = [
    cfg.StrOpt('server-endpoint',
               default=utils.env('SHAKER_SERVER_ENDPOINT'),
               required=True,
               help='Address for server connections (host:port), '
                    'defaults to env[SHAKER_SERVER_ENDPOINT].'),
    cfg.IntOpt('polling-interval',
               default=10,
               help='How frequently the agent polls server, in seconds')
]

OPENSTACK_OPTS = [
    cfg.StrOpt('os-auth-url', metavar='<auth-url>',
               default=utils.env('OS_AUTH_URL'),
               sample_default='',
               help='Authentication URL, defaults to env[OS_AUTH_URL].'),
    cfg.StrOpt('os-tenant-name', metavar='<auth-tenant-name>',
               default=utils.env('OS_TENANT_NAME'),
               sample_default='',
               help='Authentication tenant name, defaults to '
                    'env[OS_TENANT_NAME].'),
    cfg.StrOpt('os-username', metavar='<auth-username>',
               default=utils.env('OS_USERNAME'),
               sample_default='',
               help='Authentication username, defaults to env[OS_USERNAME].'),
    cfg.StrOpt('os-password', metavar='<auth-password>',
               default=utils.env('OS_PASSWORD'),
               sample_default='',
               help='Authentication password, defaults to env[OS_PASSWORD].'),
    cfg.StrOpt('os-region-name', metavar='<auth-region-name>',
               default=utils.env('OS_REGION_NAME') or 'RegionOne',
               help='Authentication region name, defaults to '
                    'env[OS_REGION_NAME].'),

    cfg.StrOpt('external-net',
               default=utils.env('SHAKER_EXTERNAL_NET'),
               help='Name or ID of external network, defaults to '
                    'env[SHAKER_EXTERNAL_NET]. If no value provided then '
                    'Shaker picks any of available external networks.'),

    cfg.StrOpt('image-name',
               default=utils.env('SHAKER_IMAGE') or 'shaker-image',
               help='Name of image to use. The default is created by '
                    'shaker-image-builder.'),
    cfg.StrOpt('flavor-name',
               default=utils.env('SHAKER_FLAVOR') or 'shaker-flavor',
               help='Name of image flavor. The default is created by '
                    'shaker-image-builder.'),
]

SERVER_OPTS = [
    cfg.StrOpt('scenario',
               default=utils.env('SHAKER_SCENARIO'),
               required=True,
               help='Scenario file name, defaults to env[SHAKER_SCENARIO].'),

    cfg.StrOpt('output',
               default=utils.env('SHAKER_OUTPUT'),
               help='File for output in JSON format, '
                    'defaults to env[SHAKER_OUTPUT].'),
]

REPORT_OPTS = [
    cfg.StrOpt('report-template',
               default=(utils.env('SHAKER_REPORT_TEMPLATE') or
                        'shaker/resources/report_template.jinja2'),
               help='Report template in Jinja format'),
    cfg.StrOpt('report',
               default=utils.env('SHAKER_REPORT'),
               help='Report file name, defaults to env[SHAKER_REPORT]. '
                    'If no value provided the report is printed to stdout.'),
]

INPUT_OPTS = [
    cfg.StrOpt('input',
               default=utils.env('SHAKER_INPUT'),
               required=True,
               help='File to read test results from, '
                    'defaults to env[SHAKER_INPUT].'),
]


AGENT_OPTS = [
    cfg.StrOpt('agent-id',
               default=utils.env('SHAKER_AGENT_ID'),
               required=True,
               help='Agent unique id, defaults to env[SHAKER_AGENT_ID].'),
]

IMAGE_BUILDER_OPTS = [
    cfg.StrOpt('image-builder-template',
               default=(utils.env('SHAKER_IMAGE_BUILDER_TEMPLATE') or
                        'shaker/resources/image_builder_template.yaml'),
               help='Heat template for the image builder.'),
]


def list_opts():
    all_opts = (COMMON_OPTS + OPENSTACK_OPTS + SERVER_OPTS + REPORT_OPTS +
                INPUT_OPTS + AGENT_OPTS + IMAGE_BUILDER_OPTS)
    yield (None, copy.deepcopy(all_opts))



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from shaker.engine import messaging
from shaker.engine import server


LOG = logging.getLogger(__name__)


class Shaker(object):
    """How to use Shaker as library

    shaker = Shaker('127.0.0.1:5999', ['the-agent'])
    res = shaker.run_program('the-agent', 'ls -al')
    """
    def __init__(self, server_endpoint, agent_ids, polling_interval=1):
        self.server_endpoint = server_endpoint
        self.polling_interval = polling_interval

        message_queue = messaging.MessageQueue(self.server_endpoint)
        self.quorum = server.Quorum(message_queue, self.polling_interval)
        self.quorum.wait_join(agent_ids)

    def _run(self, agent_id, item):
        agents = dict([(agent_id, dict(id=agent_id, mode='alone'))])

        test = {'class': 'shell'}
        test.update(item)

        execution = {'tests': [test]}
        execution_result = server.execute(self.quorum, execution, agents)

        results_per_iteration = execution_result[0]['results_per_iteration']
        results_per_agent = results_per_iteration[0]['results_per_agent']
        return dict((s['agent']['id'], s) for s in results_per_agent)

    def run_program(self, agent_id, program):
        return self._run(agent_id, {'program': program})

    def run_script(self, agent_id, script):
        return self._run(agent_id, {'script': script})



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging as std_logging
import os
import random

from oslo_config import cfg
from oslo_log import log as logging
import six


LOG = logging.getLogger(__name__)


def env(*_vars, **kwargs):
    """Returns the first environment variable set.

    If none are non-empty, defaults to '' or keyword arg default.
    """
    for v in _vars:
        value = os.environ.get(v)
        if value:
            return value
    return kwargs.get('default', None)


def validate_required_opts(conf, opts):
    # all config parameters default to ENV values, that's why standard
    # check of required options doesn't work and needs to be done manually
    for opt in opts:
        if opt.required and not conf[opt.dest]:
            raise cfg.RequiredOptError(opt.name)


def init_config_and_logging(opts):
    conf = cfg.CONF
    conf.register_cli_opts(opts)
    conf.register_opts(opts)
    logging.register_options(conf)
    logging.set_defaults()

    try:
        conf(project='shaker')
        validate_required_opts(conf, opts)
    except cfg.RequiredOptError as e:
        print('Error: %s' % e)
        conf.print_usage()
        exit(1)

    logging.setup(conf, 'shaker')
    LOG.info('Logging enabled')
    conf.log_opt_values(LOG, std_logging.DEBUG)


def read_file(file_name, base_dir=''):
    full_path = os.path.normpath(os.path.join(base_dir, file_name))
    fd = None
    try:
        fd = open(full_path)
        return fd.read()
    except IOError as e:
        LOG.error('Error reading file: %s', e)
        raise
    finally:
        if fd:
            fd.close()


def write_file(data, file_name, base_dir=''):
    full_path = os.path.normpath(os.path.join(base_dir, file_name))
    fd = None
    try:
        fd = open(full_path, 'w')
        return fd.write(data)
    except IOError as e:
        LOG.error('Error writing file: %s', e)
        raise
    finally:
        if fd:
            fd.close()


def split_address(address):
    try:
        host, port = address.split(':')
    except ValueError:
        raise ValueError('Invalid address: %s, "host:port" expected', address)
    return host, port


def read_uri(uri):
    try:
        req = six.moves.urllib.request.Request(url=uri)
        fd = six.moves.urllib.request.urlopen(req)
        raw = fd.read()
        fd.close()
        return raw
    except Exception as e:
        LOG.warn('Error "%(error)s" while reading uri %(uri)s',
                 {'error': e, 'uri': uri})


def random_string(length=6):
    return ''.join(random.sample('adefikmoprstuz', length))


def copy_dict_kv(source):
    return dict((k, v) for k, v in source.items())



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import sys

import jinja2
from oslo_config import cfg
from oslo_log import log as logging
import yaml

from shaker.engine import aggregators
from shaker.engine import config
from shaker.engine import utils


LOG = logging.getLogger(__name__)


def calculate_stats(data):
    for test_result in data['result']:
        aggregator = aggregators.get_aggregator(test_result['definition'])

        for iteration_result in test_result['results_per_iteration']:
            for agent_result in iteration_result['results_per_agent']:
                aggregator.agent_summary(agent_result)

            aggregator.iteration_summary(iteration_result)

        aggregator.test_summary(test_result)


def generate_report(report_template, report_filename, data):
    LOG.debug('Generating report, template: %s, output: %s',
              report_template, report_filename or 'stdout')

    calculate_stats(data)

    # add more filters to jinja
    jinja_env = jinja2.Environment()
    jinja_env.filters['json'] = json.dumps
    jinja_env.filters['yaml'] = functools.partial(yaml.safe_dump, indent=2,
                                                  default_flow_style=False)

    template = utils.read_file(report_template)
    compiled_template = jinja_env.from_string(template)
    rendered_template = compiled_template.render(dict(report=data))

    if report_filename:
        fd = open(report_filename, 'w')
    else:
        fd = sys.stdout

    fd.write(rendered_template)
    fd.close()
    LOG.info('Report generated')


def main():
    utils.init_config_and_logging(config.REPORT_OPTS + config.INPUT_OPTS)

    LOG.debug('Reading JSON data from: %s', cfg.CONF.input)
    report_data = json.loads(utils.read_file(cfg.CONF.input))

    generate_report(cfg.CONF.report_template, cfg.CONF.report, report_data)


if __name__ == "__main__":
    main()
